/*************************************************************************************/
/*      Copyright 2009 Barcelona Supercomputing Center                               */
/*                                                                                   */
/*      This file is part of the NANOS++ library.                                    */
/*                                                                                   */
/*      NANOS++ is free software: you can redistribute it and/or modify              */
/*      it under the terms of the GNU Lesser General Public License as published by  */
/*      the Free Software Foundation, either version 3 of the License, or            */
/*      (at your option) any later version.                                          */
/*                                                                                   */
/*      NANOS++ is distributed in the hope that it will be useful,                   */
/*      but WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                */
/*      GNU Lesser General Public License for more details.                          */
/*                                                                                   */
/*      You should have received a copy of the GNU Lesser General Public License     */
/*      along with NANOS++.  If not, see <http://www.gnu.org/licenses/>.             */
/*************************************************************************************/

#include "bf_sched.hpp"

#include "config.hpp"
#include "schedule.hpp"
#include "plugin.hpp"
#include "pollingservice.hpp"

namespace nanos {
namespace ext {

/*! \class Polling scheduler.
 * \brief Calls the polling service manager every time each worker thread
 * becomes idle.
 * It can take any scheduling policy as base.
 */
template < typename BasePolicy = BreadthFirst>
class PollingScheduler : public BasePolicy
{
   public:
      virtual ~PollingScheduler ()
      {
      }

      virtual WD * atBlock ( BaseThread *thread, WD *current )
      {
         // Do not call this class' atIdle directly.
         // We might poll a condition that is blocked by us.
         // As we already checked that condition three times at least, just give up
         // and switch context.
         return BasePolicy::atIdle(thread, false);
      }

      virtual WD * atIdle ( BaseThread *thread, int numSteal = 0 )
      {
         // Callback to registered polling services
         PollingServiceManager::poll();
         return BasePolicy::atIdle(thread,numSteal);
      }
};

bool BreadthFirst::_useStack = false;
bool BreadthFirst::_usePriority = true;
bool BreadthFirst::_useSmartPriority = false;

class PollSchedPlugin : public Plugin
{

   public:
      PollSchedPlugin() :
         Plugin( "Polling breadth first scheduling Plugin",1 )
      {
      }

      virtual void config ( Config &cfg )
      {
      }

      virtual void init() {
         sys.setDefaultSchedulePolicy(NEW PollingScheduler<>());
      }
};

} // namespace ext
} // namespace nanos

DECLARE_PLUGIN("sched-polling",nanos::ext::PollSchedPlugin);
